#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "types.h"
#include "lib.h"
#include "formula.h"
#include "command.h"
#include "line.h"
#include "label.h"
#include "value.h"
#include "string.h"
#include "variable.h"
#include "position.h"
#include "array.h"
#include "area.h"
#include "readline.h"
#include "framebuf.h"
#include "term.h"
#include "nll.h"

static int stepping = 0;

static struct spot break_spot;
static struct spot continue_spot;

static int clear_break(line_t line)
{
  if (!line || (break_spot.line == line)) {
    break_spot.line = NULL;
    break_spot.command = NULL;
  }
  return 0;
}

static int clear_continue(line_t line)
{
  if (!line || (continue_spot.line == line)) {
    continue_spot.line = NULL;
    continue_spot.command = NULL;
  }
  return 0;
}

int nll_clear_line(line_t line)
{
  clear_break(line);
  clear_continue(line);
  return 0;
}

static int get_continue(spot_t spot)
{
  if (continue_spot.line == NULL)
    return NLL_ERRCODE_NLL_NO_CONTINUE_POINT;
  spot->line    = continue_spot.line;
  spot->command = continue_spot.command;
  return 0;
}

static int set_continue(spot_t spot)
{
  continue_spot.line    = spot->line;
  continue_spot.command = spot->command;
  return 0;
}

int nll_init(void)
{
  nll_clear_line(NULL);
  variable_init();
  value_init();
  string_init();
  position_init();
  label_init();
  formula_init();
  command_init();
  line_init();
  array_init();
  area_init();
  framebuf_init();
  return 0;
}

static int _get_word(char **pp, const char *separator, char *buffer, int size)
{
  char c, *p;

  while (isspace(**pp))
    (*pp)++;

  p = buffer;

  while (1) {
    c = **pp;
    if (c == '\0') {
      break;
    } else if (strchr(separator, c)) {
      break;
    }
    **pp = toupper(c);
    if (p + 1 >= buffer + size)
      return NLL_ERRCODE_NLL_WORD_TOO_LONG;
    *(p++) = **pp;
    (*pp)++;
  }

  *p = '\0';

  return 0;
}

static int get_word(char **pp, char *buffer, int size)
{
  return _get_word(pp, ", :;(\t\n\r", buffer, size);
}

static int proc_let(command_arg_t args)
{
  int r;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, NULL)) < 0)
      return r;
  }

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, NULL)) < 0)
      return r;
    if (args[0].type == COMMAND_ARG_FORMULA) {
      if ((r = value_copy_values(args[0].arg.formula.element->value, args[1].arg.formula.element->value)) < 0)
	return r;
    }
  }

  return 0;
}

static int proc_print(command_arg_t args)
{
  int r;
  integer_t x = -1, y = -1, c = -1;
  value_t value;

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &x)) < 0)
	return r;
    }
  }

  if (args[2].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[2].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &y)) < 0)
	return r;
    }
  }

  if ((x >= 0) || (y >= 0))
    framebuf_move_cursol(x, y);

  if (args[3].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[3].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &c)) < 0)
	return r;
      framebuf_setcolor(c);
    }
  }

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      value_print(value);
    }
  }

  if ((x < 0) || (y < 0))
    framebuf_puts("\n");

  return 0;
}

static int proc_go(spot_t spot, command_arg_t args)
{
  int r;

  if ((args[0].type != COMMAND_ARG_LABEL) ||
      !args[0].arg.label.label->spot.line)
    return NLL_ERRCODE_LABEL_NOT_FOUND;

  spot->line    = args[0].arg.label.label->spot.line;
  spot->command = args[0].arg.label.label->spot.command;

  if ((r = formula_proc(args[0].arg.label.args, NULL)) < 0)
    return r;
  if ((r = formula_proc(spot->command->args[0].arg.label.args, NULL)) < 0)
    return r;
  if (spot->command->args[0].arg.label.args && args[0].arg.label.args) {
    if ((r = value_copy_values(spot->command->args[0].arg.label.args->value,
			       args[0].arg.label.args->value)) < 0)
      return r;
  }

  return 0;
}

static int proc_gonext(spot_t spot, line_t line, command_arg_t args)
{
  int r;
  integer_t n = 1, n2 = 0;
  value_t value;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &n)) < 0)
	return r;
    }
  }

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &n2)) < 0)
	return r;
    }
  }

  for (; n > 0; n--) {
    if (line)
      line = line->next;
  }

  spot->line = line;
  if (spot->line) {
    for (spot->command = spot->line->commands; n2 > 0; n2--) {
      if (spot->command)
	spot->command = spot->command->next;
    }
  }

  return 0;
}

static int proc_loop(spot_t spot, command_arg_t args)
{
  int r;
  value_t value;

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, NULL)) < 0)
      return r;
  }

  if (args[2].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[2].arg.formula.element, NULL)) < 0)
      return r;
  }

  if (args[3].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[3].arg.formula.element, NULL)) < 0)
      return r;
  }

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (args[2].type == COMMAND_ARG_FORMULA) {
      if ((r = value_copy_values(args[0].arg.formula.element->value,
				 args[2].arg.formula.element->value)) < 0)
	return r;
    } else {
      if (value) {
	if ((r = value_set_integer(value, 0)) < 0)
	  return r;
      }
    }
  }

  if ((r = position_push(POSITION_TYPE_LOOP, spot, NULL)) < 0)
    return r;

  return 0;
}

static int proc_for(spot_t spot, command_arg_t args)
{
  int r;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, NULL)) < 0)
      return r;
  }

  if ((r = position_push(POSITION_TYPE_FOR, spot, NULL)) < 0)
    return r;

  return 0;
}

static int proc_loopend(spot_t spot, command_arg_t args)
{
  int r;
  integer_t integer = 0, count = 10, start = 0, step = 1, end = 1;
  struct spot s;
  value_t value = NULL;

  if ((r = position_pop(POSITION_TYPE_LOOP, &s, NULL)) < 0)
    return r;

  if (s.line == NULL)
    return NLL_ERRCODE_NLL_NO_CONTINUE_POINT;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
  }

  args = s.command->args;

  if (value) {
    if ((args[0].type != COMMAND_ARG_FORMULA) ||
	!value_is_same(value, args[0].arg.formula.element->value))
      return NLL_ERRCODE_POSITION_DIFFERENT_VARIABLE;
  }

  value = NULL;
  if (args[0].type == COMMAND_ARG_FORMULA)
    value = args[0].arg.formula.element->value;

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = value_get_integer(args[1].arg.formula.element->value, &count)) < 0)
      return r;
  }
  if (args[2].type == COMMAND_ARG_FORMULA) {
    if ((r = value_get_integer(args[2].arg.formula.element->value, &start)) < 0)
      return r;
  }
  if (args[3].type == COMMAND_ARG_FORMULA) {
    if ((r = value_get_integer(args[3].arg.formula.element->value, &step)) < 0)
      return r;
  }

  if (value) {
    if ((r = value_get_integer(value, &integer)) < 0)
      return r;

    integer += step;

    if ((r = value_set_integer(value, integer)) < 0)
      return r;

    end = start + step * count;
  }

  if (integer != end) {
    if ((r = position_push(POSITION_TYPE_LOOP, &s, NULL)) < 0)
      return r;
    spot->line    = s.line;
    spot->command = s.command;
    return 1;
  }

  return 0;
}

static int proc_next(spot_t spot, command_arg_t args)
{
  int r;
  struct spot s;
  value_t value = NULL;

  if ((r = position_pop(POSITION_TYPE_FOR, &s, NULL)) < 0)
    return r;

  if (s.line == NULL)
    return NLL_ERRCODE_NLL_NO_CONTINUE_POINT;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, NULL)) < 0)
      return r;
  }

  args = s.command->args;

  if (args[2].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[2].arg.formula.element, NULL)) < 0)
      return r;
  }

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, &value)) < 0)
      return r;
  }

  if (value) {
    if ((r = value_is_nzero(value)) < 0)
      return r;
    if (!r)
      value = NULL;
  }

  if (value) {
    if ((r = position_push(POSITION_TYPE_FOR, &s, NULL)) < 0)
      return r;
    spot->line    = s.line;
    spot->command = s.command;
    return 1;
  }

  return 0;
}

static int proc_discard(command_arg_t args)
{
  int r;
  integer_t n = 1;
  value_t value;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &n)) < 0)
	return r;
    }
  }

  for (; n > 0; n--) {
    if ((r = position_pop(POSITION_TYPE_NONE, NULL, NULL)) < 0)
      return r;
  }

  return 0;
}

static int make_array(value_t value, int *n)
{
  int r, i;
  array_t array;

  if (*n == 0)
    return 0;
  if ((r = array_get(&array, *n)) < 0)
    return r;
  for (i = 0; i < *n; i++) {
    if ((r = make_array(array[i].value, n + 1)) < 0)
      return r;
  }
  if ((r = value_set_array(value, *n, array)) < 0)
    return r;

  return 0;
}

static int proc_dim(command_arg_t args)
{
  int r, i, n[ARRAY_DIM_MAXNUM + 1];
  integer_t integer;
  value_t value;

  for (i = 0; i < ARRAY_DIM_MAXNUM; i++) {
    n[i] = 0;
    if (args[i + 1].type == COMMAND_ARG_FORMULA) {
      if ((r = formula_proc(args[i + 1].arg.formula.element, &value)) < 0)
	return r;
      if (value) {
	if ((r = value_get_integer(value, &integer)) < 0)
	  return r;
	n[i] = integer;
      } else {
	n[i] = 1;
      }
    }
  }
  n[i] = 0;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = make_array(value, n)) < 0)
	return r;
    }
  }

  return 0;
}

static int proc_memory(command_arg_t args)
{
  int r;
  integer_t integer;
  value_t value;
  area_t area;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, NULL)) < 0)
      return r;
  }

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, &value)) < 0)
      return r;
    if (args[0].type == COMMAND_ARG_FORMULA) {
      if ((r = value_get_integer(value, &integer)) < 0)
	return r;
      if ((r = area_alloc(&area, integer)) < 0)
	return r;
      if ((r = value_set_area(args[0].arg.formula.element->value, area)) < 0)
	return r;
    }
  }

  return 0;
}

static int proc_run(spot_t spot, command_arg_t args)
{
  int r;
  integer_t integer = 1;
  value_t value;
  line_t line = NULL;
  command_t command = NULL;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &integer)) < 0)
	return r;
    }
  }

  if ((r = line_get(integer, &line)) < 0)
    return r;

#if 0
  if (!line)
    line = line_head();
#endif

  if (line)
    command = line->commands;

  spot->line    = line;
  spot->command = command;

  return 0;
}

static int proc_break(command_arg_t args)
{
  int r;
  integer_t integer0 = 0, integer1 = 0;
  value_t value;
  line_t line = NULL;
  command_t command;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &integer0)) < 0)
	return r;
    }
    if ((r = line_get(integer0, &line)) < 0)
      return r;
  }

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &integer1)) < 0)
	return r;
    }
  }

  if (!line) {
    clear_break(NULL);
  } else {
    for (command = line->commands; command && (integer1 > 0); integer1--) {
      command = command->next;
    }
    break_spot.line    = line;
    break_spot.command = command;
  }

  return 0;
}

static int proc_step(command_arg_t args)
{
  int r;
  integer_t integer = 1;
  value_t value;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &integer)) < 0)
	return r;
    }
  }

  stepping = integer;

  return 0;
}

static int proc_new(spot_t spot, command_arg_t args)
{
  int r;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, NULL)) < 0)
      return r;
    if ((r = value_clear_value(args[0].arg.formula.element->value)) < 0)
      return r;
    return 0;
  }

  return 1;
}

static int proc_list(command_arg_t args)
{
  int r;
  integer_t integer0 = -1, integer1 = -1;
  value_t value;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &integer0)) < 0)
	return r;
    }
  }

  integer1 = integer0;

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &integer1)) < 0)
	return r;
    }
  }

  line_print(integer0, integer1);
  fflush(stdout);

  return 0;
}

static int proc_edit(command_arg_t args)
{
  int r;
  integer_t integer = 0;
  value_t value;
  line_t line;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &integer)) < 0)
	return r;
    }
  }

  if ((r = line_get(integer, &line)) < 0)
    return r;

  if (line) {
    nll_edit_line(integer, line);
  }

  return 0;
}

static int proc_wait(command_arg_t args)
{
  int r;
  integer_t integer = 100;
  value_t value;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &integer)) < 0)
	return r;
    }
  }

  usleep(integer * 10000);

  return 0;
}

static int proc_dump(command_arg_t args)
{
  char *word = NULL;
  int all = 0, variable = 0, string = 0, label = 0, line = 0;

  if (args[0].type == COMMAND_ARG_WORD)
    word = args[0].arg.word.word;

  if (!word || !strcmp(word, "ALL") || !strcmp(word, "A")) {
    all = 1;
  } else {
    if (!strcmp(word, "VARIABLE") || !strcmp(word, "VAL") || !strcmp(word, "V"))
      variable = 1;
    if (!strcmp(word, "STRING") || !strcmp(word, "STR") || !strcmp(word, "S"))
      string = 1;
    if (!strcmp(word, "LABEL") || !strcmp(word, "L"))
      label = 1;
    if (!strcmp(word, "LINE") || !strcmp(word, "LN"))
      line = 1;
  }

  if (all || variable) variable_dump(NULL);
  if (all || string) string_dump(NULL);
  if (all || label) label_dump(NULL);
  if (all || line) line_dump(NULL);

  return 0;
}

static void set_errcode(command_t command, int r)
{
  if (command)
    command->errcode = r;
}

#define ERRBREAK(r, command, param) \
	do { NLL_ERRPRINT((r), (param)); set_errcode((command), (r)); goto brk; } while (0)

int nll_parse(line_t line)
{
  char *p, *tail, *buffer, *pc;
#define ARG_LEN 256
  char arg0[ARG_LEN];
  int r, count = 0;
  command_t command = NULL, *tailp;
  struct spot s;

  buffer = line->buffer;
  p = buffer;
  tail = p + strlen(p);

  tailp = &line->commands;

  while (p < tail) {
    if (isspace(*p))
      goto cont;

    pc = p;

    if ((r = command_alloc(&command)) < 0) ERRBREAK(r, command, pc);
    command->line = pc;

    if (*p == '.') {
      p++;
      if ((r = get_word(&p, arg0, ARG_LEN)) < 0)
	ERRBREAK(r, command, pc);
      if (arg0[0] == '\0')
	ERRBREAK(NLL_ERRCODE_LABEL_INVALID_NAME, command, pc);

      if ((r = command_get(COMMAND_LABEL, &command->ope)) < 0)
	ERRBREAK(r, command, pc);
      s.line    = line;
      s.command = command;
      if ((r = label_set(arg0, &s, &command->args[0].arg.label.label)) < 0)
	ERRBREAK(r, command, pc);
      command->args[0].type = COMMAND_ARG_LABEL;

      while (isspace(*p))
	p++;
      if (*p == ':') {
	p++;
	if ((r = formula_parse(p, &p, &command->args[0].arg.label.args, ";:")) < 0)
	  ERRBREAK(r, command, pc);
      }

      goto cont;
    }

    if ((r = command_search(p, &p, &command->ope)) < 0) {
      if ((r = formula_parse(p, &p, &command->prefix, NULL)) < 0)
	ERRBREAK(r, command, pc);
      if ((r = command_get(COMMAND_NONE, &command->ope)) < 0)
	ERRBREAK(r, command, pc);
      if (*p == ',') {
	p++;
	while (isspace(*p))
	  p++;
	if ((r = command_search(p, &p, &command->ope)) < 0)
	  ERRBREAK(r, command, pc);
      }
    }

    switch (command->ope->type) {
    case COMMAND_NONE:
      break;

    case COMMAND_COMMENT:
      goto brk;

    case COMMAND_END:
    case COMMAND_EXIT:
    case COMMAND_STOP:
    case COMMAND_CONTINUE:
      break;

    case COMMAND_GOTO:
      if ((r = get_word(&p, arg0, ARG_LEN)) < 0) ERRBREAK(r, command, pc);
      if (arg0[0]) {
	if ((r = label_set(arg0, NULL, &command->args[0].arg.label.label)) < 0) ERRBREAK(r, command, pc);
	command->args[0].type = COMMAND_ARG_LABEL;
	while (isspace(*p))
	  p++;
	if (*p == ':') {
	  p++;
	  if ((r = formula_parse(p, &p, &command->args[0].arg.label.args, NULL)) < 0)
	    ERRBREAK(r, command, pc);
	}
      }
      break;

    case COMMAND_GOSUB:
      if ((r = get_word(&p, arg0, ARG_LEN)) < 0) ERRBREAK(r, command, pc);
      if (arg0[0]) {
	if ((r = label_set(arg0, NULL, &command->args[0].arg.label.label)) < 0) ERRBREAK(r, command, pc);
	command->args[0].type = COMMAND_ARG_LABEL;
	while (isspace(*p))
	  p++;
	if (*p == ':') {
	  p++;
	  if ((r = formula_parse(p, &p, &command->args[0].arg.label.args, NULL)) < 0)
	    ERRBREAK(r, command, pc);
	}
      }
      if (*p == ',') p++; else break;
      if ((r = formula_parse(p, &p, &command->args[1].arg.formula.element, NULL)) < 0)
	ERRBREAK(r, command, pc);
      if (command->args[1].arg.formula.element)
	command->args[1].type = COMMAND_ARG_FORMULA;
      break;

    case COMMAND_LET:
    case COMMAND_GONEXT:
    case COMMAND_LOOP:
    case COMMAND_FOR:
    case COMMAND_DIM:
    case COMMAND_MEMORY:
    case COMMAND_LOOPEND:
    case COMMAND_NEXT:
    case COMMAND_DISCARD:
    case COMMAND_RETURN:
    case COMMAND_PRINT:
    case COMMAND_RUN:
    case COMMAND_STEP:
    case COMMAND_EDIT:
    case COMMAND_WAIT:
    case COMMAND_BREAK:
    case COMMAND_NEW:
    case COMMAND_LIST:
      if ((r = formula_parse(p, &p, &command->args[0].arg.formula.element, NULL)) < 0)
	ERRBREAK(r, command, pc);
      if (command->args[0].arg.formula.element)
	command->args[0].type = COMMAND_ARG_FORMULA;
      if (*p == ',') p++; else break;
      if ((r = formula_parse(p, &p, &command->args[1].arg.formula.element, NULL)) < 0)
	ERRBREAK(r, command, pc);
      if (command->args[1].arg.formula.element)
	command->args[1].type = COMMAND_ARG_FORMULA;
      if (*p == ',') p++; else break;
      if ((r = formula_parse(p, &p, &command->args[2].arg.formula.element, NULL)) < 0)
	ERRBREAK(r, command, pc);
      if (command->args[2].arg.formula.element)
	command->args[2].type = COMMAND_ARG_FORMULA;
      if (*p == ',') p++; else break;
      if ((r = formula_parse(p, &p, &command->args[3].arg.formula.element, NULL)) < 0)
	ERRBREAK(r, command, pc);
      if (command->args[3].arg.formula.element)
	command->args[3].type = COMMAND_ARG_FORMULA;
      break;

    case COMMAND_DUMP:
      if ((r = get_word(&p, arg0, ARG_LEN)) < 0) ERRBREAK(r, command, pc);
      if (arg0[0]) {
	if (strlen(arg0) > COMMAND_ARG_WORD_MAXLEN)
	  ERRBREAK(NLL_ERRCODE_COMMAND_TOO_LONG, command, pc);
	strcpy(command->args[0].arg.word.word, arg0);
	command->args[0].type = COMMAND_ARG_WORD;
      }
      break;

    case COMMAND_UNKNOWN:
    default:
      ERRBREAK(NLL_ERRCODE_COMMAND_NOT_FOUND, command, pc);
    }

    cont:
    if (command) {
      command->next = *tailp;
      *tailp = command;
      tailp = &(command->next);
      command = NULL;
      count++;
    }

    if (*p)
      p++;
  }

  r = count;

brk:
  if (command) {
    if (!command->errcode) {
      command_free(command);
    } else {
      command_clear(command);
      command->next = *tailp;
      *tailp = command;
      tailp = &(command->next);
    }
  }

  return r;
}

static int _nll_run(spot_t spot)
{
  line_t line;
  command_t command;
  int r, count = 0;
  struct spot s;
  value_t value;

  line    = spot->line;
  command = spot->command;

  for (; command; command = command->next) {
    if (nll_is_finished()) {
      printf("\nBreak at: %s\n", command->line);
      goto brk;
    }

    if (break_spot.line && (line_is_exist(line) > 0)) {
      if ((line == break_spot.line) && (command == break_spot.command)) {
	clear_break(line);
	printf("\nBreakpoint at: %s\n", command->line);
	goto brk;
      }
    }

    if (stepping && (line_is_exist(line) > 0)) {
      stepping--;
      if (!stepping) {
	printf("\nStep break at: %s\n", command->line);
	goto brk;
      }
    }

    if (command->errcode) {
      NLL_ERRPRINT(command->errcode, command->line);
      goto brk;
    }

    count++;

    if (command->prefix) {
      if ((r = formula_proc(command->prefix, &value)) < 0)
	ERRBREAK(r, NULL, command->line);
      if (!value) {
	r = 0;
      } else {
	if ((r = value_is_nzero(value)) < 0)
	  ERRBREAK(r, NULL, command->line);
      }
      if (!r)
	continue;
    }

    switch (command->ope->type) {
    case COMMAND_END:
      goto end;

    case COMMAND_EXIT:
      goto ext;

    case COMMAND_LET:
      if ((r = proc_let(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_PRINT:
      if ((r = proc_print(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_GOTO:
      if ((r = proc_go(spot, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      goto next;

    case COMMAND_GONEXT:
      if ((r = proc_gonext(spot, line, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      goto next;

    case COMMAND_GOSUB:
      if ((r = proc_go(spot, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      s.line    = line;
      s.command = command;
      if ((r = position_push(POSITION_TYPE_GOSUB, &s, NULL)) < 0)
	ERRBREAK(r, NULL, command->line);
      goto next;

    case COMMAND_RETURN:
      if (command->args[0].type == COMMAND_ARG_FORMULA) {
	if ((r = formula_proc(command->args[0].arg.formula.element, NULL)) < 0)
	  ERRBREAK(r, NULL, command->line);
      }
      if ((r = position_pop(POSITION_TYPE_GOSUB, spot, NULL)) < 0) {
	if ((r = position_pop(POSITION_TYPE_LABEL, spot, NULL)) < 0)
	  ERRBREAK(r, NULL, command->line);
	value = NULL;
	if (command->args[0].type == COMMAND_ARG_FORMULA)
	  value = command->args[0].arg.formula.element->value;
	if ((r = position_push(POSITION_TYPE_LABEL, NULL, value)) < 0)
	  return r;
	goto end;
      }
      if (spot->line == NULL)
	ERRBREAK(NLL_ERRCODE_NLL_NO_CONTINUE_POINT, NULL, command->line);
      if ((command->args[0].type == COMMAND_ARG_FORMULA) &&
	  (spot->command->args[1].type == COMMAND_ARG_FORMULA)) {
	if ((r = formula_proc(spot->command->args[1].arg.formula.element, NULL)) < 0)
	  ERRBREAK(r, NULL, command->line);
	if ((r = value_copy_values(spot->command->args[1].arg.formula.element->value,
				   command->args[0].arg.formula.element->value)) < 0)
	  ERRBREAK(r, NULL, command->line);
      }
      spot->command = spot->command->next;
      goto next;

    case COMMAND_LOOP:
      s.line    = line;
      s.command = command;
      if ((r = proc_loop(&s, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_FOR:
      s.line    = line;
      s.command = command;
      if ((r = proc_for(&s, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_LOOPEND:
      if ((r = proc_loopend(spot, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      if (!r)
	break;
      spot->command = spot->command->next;
      goto next;

    case COMMAND_NEXT:
      if ((r = proc_next(spot, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      if (!r)
	break;
      spot->command = spot->command->next;
      goto next;

    case COMMAND_DISCARD:
      if ((r = proc_discard(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_DIM:
      if ((r = proc_dim(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_MEMORY:
      if ((r = proc_memory(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_RUN:
      if ((r = proc_run(spot, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      line_clean();
      variable_clear();
      position_init();
      clear_continue(NULL);
      stepping = 0;
      goto next;

    case COMMAND_STOP:
      printf("\nStop at: %s\n", command->line);
      command = command->next;
      goto brk;

    case COMMAND_BREAK:
      if ((r = proc_break(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_STEP:
      if ((r = proc_step(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      if (stepping) stepping++;
      /* fall through */

    case COMMAND_CONTINUE:
      if ((r = get_continue(spot)) < 0)
	ERRBREAK(r, NULL, command->line);
      clear_continue(NULL);
      goto next;

    case COMMAND_NEW:
      if ((r = proc_new(spot, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      if (!r)
	break;
      goto init;

    case COMMAND_LIST:
      if ((r = proc_list(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_EDIT:
      if ((r = proc_edit(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_WAIT:
      if ((r = proc_wait(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_DUMP:
      if ((r = proc_dump(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_NONE:
    case COMMAND_LABEL:
    case COMMAND_UNKNOWN:
    default:
      break;
    }
  }

  spot->line = line->next;
  if (spot->line)
    spot->command = spot->line->commands;

next:
  return 0;

ext:
  spot->line = NULL;
  return -1;

end:
  spot->line = NULL;
  return 0;

init:
  spot->line = NULL;
  return 1;

brk:
  s.line    = line;
  s.command = command;
  if ((r = set_continue(&s)) < 0)
    ERRBREAK(r, NULL, command->line);
  spot->line = NULL;
  return 0;
}

int nll_run(spot_t spot)
{
  int r = 0;

  while (spot->line) {
    r = _nll_run(spot);
    fflush(stdout);
  }

  return r;
}

int nll_start(line_t line, command_t command)
{
  struct spot spot;
  int r;

  nll_finished_clear();

  spot.line = line;

  if (!line)
    spot.command = NULL;
  else
    spot.command = command ? command : line->commands;

  nll_term_start();

  r = nll_run(&spot);

  nll_term_restore();

  return r;
}
